# PaCMAP-GPU: GPU-Accelerated Pairwise Controlled Manifold Approximation

[![Python](https://img.shields.io/badge/Python-3.10%2B-blue)](https://www.python.org/)
[![CUDA](https://img.shields.io/badge/CUDA-12.0%2B-green)](https://developer.nvidia.com/cuda-toolkit)
[![PyTorch](https://img.shields.io/badge/PyTorch-2.0%2B-red)](https://pytorch.org/)
[![RAPIDS](https://img.shields.io/badge/RAPIDS-24.0%2B-purple)](https://rapids.ai/)
[![License](https://img.shields.io/badge/License-Apache%202.0-blue.svg)](https://opensource.org/licenses/Apache-2.0)

**PaCMAP-GPU** is a high-performance GPU implementation of Pairwise Controlled Manifold Approximation and Projection (PaCMAP), providing dramatic speedups for large-scale dimensionality reduction and manifold learning tasks.

## 🚀 Key Features

- **GPU Acceleration**: Native CUDA implementation using PyTorch and CUDA graphs for maximum performance
- **Massive Speedups**: 5-50x faster than CPU implementations depending on dataset size
- **Memory Efficient**: Optimized GPU memory usage with automatic cleanup
- **Drop-in Replacement**: Compatible API with sklearn-style transformers
- **Three-Phase Training**: Standard PaCMAP phases plus LocalMAP variant
- **Large Dataset Support**: Handle datasets with millions of points efficiently
- **CUDA Graph Optimization**: Minimal kernel launch overhead

## 📋 Requirements

### Hardware Requirements
- NVIDIA GPU with Compute Capability 6.0+ (Pascal architecture or newer)
- 4GB+ GPU memory (8GB+ recommended for large datasets)
- CUDA 12.0 or higher

### Software Dependencies
Core dependencies (automatically installed):
- PyTorch >= 2.0.0
- NumPy >= 1.21.0, < 2.0.0
- SciPy >= 1.7.0
- scikit-learn >= 1.0.0
- pandas >= 1.3.0
- matplotlib >= 3.5.0

GPU acceleration (required):
- cupy-cuda12x >= 12.0.0
- cuml-cu12 >= 24.0.0
- cudf-cu12 >= 24.0.0
- faiss-gpu >= 1.7.0

**Important: CUDA-enabled PyTorch Required**

> By default, `pip install pacmap-gpu` will install the CPU-only version of PyTorch.  
> **To use GPU acceleration, you must manually install the CUDA-enabled PyTorch wheel** that matches your system and CUDA version.  
> See [PyTorch Get Started](https://pytorch.org/get-started/locally/) for installation instructions.

**Example:**
```bash
# Example for CUDA 12.1
pip install torch==2.2.2+cu121 torchvision==0.17.2+cu121 torchaudio==2.2.2+cu121 \
  --index-url https://download.pytorch.org/whl/cu121
```

## 🔧 Installation

### Option 1: Install from PyPI (Recommended)

```bash
# Install PaCMAP-GPU with all dependencies
pip install pacmap-gpu

# Verify installation
pacmap-gpu-check
```

**Note:** PaCMAP-GPU requires CUDA-capable hardware and GPU libraries. All necessary RAPIDS and CUDA dependencies are automatically installed.

### Option 2: Install from Source

```bash
# Clone the repository
git clone https://github.com/AntonioRoye/PaCMAP-GPU.git
cd PaCMAP-GPU

# Create conda environment with RAPIDS and all dependencies (recommended for development)
conda create -n pacmap-gpu -c rapidsai -c conda-forge \
    cuml-cu12 cudf-cu12 cupy-cuda12x faiss-gpu \
    python=3.10 pytorch pytorch-cuda=12.1 -c pytorch -c nvidia

# Activate environment
conda activate pacmap-gpu

# Install package in development mode
pip install -e .
```

### Alternative: Using requirements file

```bash
# Clone repository
git clone https://github.com/AntonioRoye/PaCMAP-GPU.git
cd PaCMAP-GPU

# Create virtual environment
python -m venv venv
source venv/bin/activate  # On Windows: venv\Scripts\activate

# Install CUDA 12 dependencies
pip install -r requirements-cuda12.txt

# Install package
pip install -e .
```

### Verify Installation

```bash
# Check installation
pacmap-gpu-check

# Or in Python
python -c "from pacmap_gpu import PaCMAPGPU, check_gpu_availability; print('GPU Available:', check_gpu_availability())"
```

## 🎯 Quick Start

### Basic Usage

```python
import numpy as np
from pacmap_gpu import PaCMAPGPU
from sklearn.datasets import make_swiss_roll

# Generate sample data
X, color = make_swiss_roll(n_samples=2000, noise=0.1, random_state=42)

# Create PaCMAP-GPU instance
reducer = PaCMAPGPU(
    n_components=2,
    n_neighbors=10,
    num_iters=450,
    random_state=42,
    verbose=True
)

# Fit and transform the data
X_embedded = reducer.fit_transform(X)
print(f"Embedded shape: {X_embedded.shape}")
```

### Advanced Configuration

```python
# High-performance configuration for large datasets
reducer = PaCMAPGPU(
    n_components=2,
    n_neighbors=15,
    MN_ratio=0.5,
    FP_ratio=2.0,
    num_iters=450,
    batch_size=8192,
    init_method="pca",
    optimizer="adam",
    lr=0.01,
    low_dist_thres=1.0,  # LocalMAP parameter
    device="cuda",
    verbose=True,
    random_state=42
)

X_embedded = reducer.fit_transform(X)
```

### Visualization Example

```python
import matplotlib.pyplot as plt
from sklearn.datasets import load_digits

# Load digits dataset
digits = load_digits()
X, y = digits.data, digits.target

# Apply PaCMAP-GPU
reducer = PaCMAPGPU(n_components=2, verbose=True, random_state=42)
X_embedded = reducer.fit_transform(X)

# Plot results
plt.figure(figsize=(10, 8))
scatter = plt.scatter(X_embedded[:, 0], X_embedded[:, 1], 
                     c=y, cmap='tab10', alpha=0.7)
plt.colorbar(scatter)
plt.title('Digits Dataset - PaCMAP-GPU Embedding')
plt.xlabel('Component 1')
plt.ylabel('Component 2')
plt.show()
```

## 📊 Performance Benchmarks

### Speedup Comparison

Performance varies based on dataset size and GPU hardware. Below are typical dataset configurations for benchmarking:

| Dataset Size | Features | Use Case | Notes |
|-------------|----------|----------|-------|
| 1K samples  | 50       | Quick testing | May show minimal speedup due to GPU overhead |
| 5K samples  | 100      | Standard evaluation | ~2-5x speedup typical |  
| 10K samples | 200      | Medium datasets | ~5-15x speedup expected |
| 50K samples | 500      | Large-scale analysis | ~10-50x speedup possible |

*Actual performance depends on GPU model, CUDA version, and data characteristics*

**Run your own benchmarks:**
```bash
pytest tests/test_benchmarks.py -v -m benchmark
```

### Memory Efficiency

- **PaCMAP-GPU**: ~2-4GB GPU memory for 50K samples (with CUDA graph optimizations)
- **Traditional Methods**: ~4-8GB RAM for 50K samples (varies by dimensionality)  
- **Automatic Cleanup**: Memory pools and CUDA graph optimization prevent OOM errors

## 🔍 API Reference

### PaCMAPGPU Class

```python
PaCMAPGPU(
    n_components=2,          # Output dimensions
    n_neighbors=10,          # Neighbors for local structure
    MN_ratio=0.5,           # Mid-near pairs ratio
    FP_ratio=2.0,           # Further pairs ratio
    num_iters=450,          # Total optimization iterations
    batch_size=8192,        # Batch size for stochastic optimization
    init_method="pca",      # Initialization method
    optimizer="adam",       # Optimizer type
    lr=0.01,               # Learning rate
    device="cuda",         # CUDA device
    low_dist_thres=1.0,    # LocalMAP distance threshold
    random_state=None,     # Random seed
    verbose=False          # Print progress
)
```

#### Key Methods

- `fit_transform(X)`: Fit and return embedding
- `transform(X)`: Transform new data (not implemented)

#### Attributes

- `embedding_`: Final embedding coordinates
- `is_fitted`: Whether model has been fitted
- `n_samples`: Number of training samples

### Utility Functions

```python
from pacmap_gpu import check_gpu_availability, get_gpu_info, clear_gpu_memory

# Check GPU status
gpu_available = check_gpu_availability()
gpu_info = get_gpu_info()

# Memory management
clear_gpu_memory()

# Installation verification
from pacmap_gpu import verify_installation, print_installation_report
print_installation_report()
```

## ⚡ Performance Tips

### Dataset Size Guidelines
- **Small datasets (< 1K)**: GPU overhead may impact performance, but still required
- **Medium datasets (1K-10K)**: GPU shows significant speedup over traditional CPU methods  
- **Large datasets (10K+)**: Maximum GPU performance advantage

### Memory Optimization
```python
from pacmap_gpu import clear_gpu_memory

# Clear memory between runs
clear_gpu_memory()

# Use efficient data types
X = X.astype(np.float32)  # Use float32 instead of float64
```

### Hardware Optimization
- Use modern GPUs with Tensor Cores (RTX series, V100, A100)
- Ensure sufficient GPU memory (8GB+ for datasets > 50K samples)
- Use CUDA 12.0+ for best performance

## 🧪 Running Tests and Benchmarks

```bash
# Install development dependencies
pip install -e ".[dev]"

# Run basic tests
pytest tests/ -v

# Run GPU tests (requires GPU)
pytest tests/ -v -m gpu

# Run benchmarks
pytest tests/test_benchmarks.py -v -m benchmark

# Run without slow tests
pytest tests/ -v -m "not slow"
```

### Example Benchmark Output

```bash
pytest tests/test_benchmarks.py::TestPerformanceBenchmarks::test_gpu_vs_cpu_performance -v -s

=== MEDIUM Dataset Benchmark ===
Dataset shape: (2000, 50)
GPU time: 1.8432s
CPU time: 14.2156s
Speedup: 7.71x
```

## 🐛 Troubleshooting

### Common Issues

1. **CUDA Not Found**
   ```bash
   # Check CUDA installation
   nvidia-smi
   python -c "import torch; print(torch.cuda.is_available())"
   ```

2. **Out of Memory Error**
   ```python
   # Reduce batch size or data size
   reducer = PaCMAPGPU(batch_size=2048, n_neighbors=5)
   ```

3. **Import Errors**
   ```bash
   # Verify installation
   pacmap-gpu-check
   
   # Reinstall if needed
   pip install --force-reinstall pacmap-gpu
   ```

### Performance Issues

- **Slow on small datasets**: GPU overhead exists, but GPU is still required for functionality
- **GPU underutilized**: Increase dataset size or batch parameters
- **Memory warnings**: Monitor GPU memory and clear caches regularly

## 🤝 Contributing

We welcome contributions! Please follow these steps:

1. Fork the repository
2. Create a feature branch (`git checkout -b feature/amazing-feature`)
3. Make changes and add tests
4. Run the test suite (`pytest tests/`)
5. Commit changes (`git commit -m 'Add amazing feature'`)
6. Push to branch (`git push origin feature/amazing-feature`)
7. Open a Pull Request

### Development Setup

```bash
# Clone repository
git clone https://github.com/AntonioRoye/PaCMAP-GPU.git
cd PaCMAP-GPU

# Set up conda environment with all dependencies (recommended)
conda create -n pacmap-gpu-dev -c rapidsai -c conda-forge \
    cuml-cu12 cudf-cu12 cupy-cuda12x faiss-gpu \
    python=3.10 pytorch pytorch-cuda=12.1 -c pytorch -c nvidia
conda activate pacmap-gpu-dev

# Install package and development tools
pip install -e ".[dev]"

# Or alternatively, install from requirements file + dev extras
# pip install -r requirements-cuda12.txt
# pip install -e ".[dev]"

# Run tests
pytest tests/

# Run GPU-specific tests (requires CUDA)
pytest tests/ -m gpu

# Format code
black source/

# Type checking
mypy source/
```

## 📚 Citation

If you use PaCMAP-GPU in your research, please cite:

```bibtex
@article{JMLR:v22:20-1061,
  author  = {Yingfan Wang and Haiyang Huang and Cynthia Rudin and Yaron Shaposhnik},
  title   = {Understanding How Dimension Reduction Tools Work: An Empirical Approach to Deciphering t-SNE, UMAP, TriMap, and PaCMAP for Data Visualization},
  journal = {Journal of Machine Learning Research},
  year    = {2021},
  volume  = {22},
  number  = {201},
  pages   = {1-73},
  url     = {http://jmlr.org/papers/v22/20-1061.html}
}

@software{pacmap_gpu2025,
  title={PaCMAP-GPU: GPU-Accelerated Pairwise Controlled Manifold Approximation},
  author={Antonio Roye-Azar},
  year={2025},
  url={https://github.com/AntonioRoye/PaCMAP-GPU}
}
```

## 📄 License

This project is licensed under the Apache License 2.0 - see the [LICENSE](LICENSE) file for details.

## 🔗 Links

- **Original PaCMAP**: [https://github.com/YingfanWang/PaCMAP](https://github.com/YingfanWang/PaCMAP)
- **PyTorch**: [https://pytorch.org/](https://pytorch.org/)
- **RAPIDS cuML**: [https://docs.rapids.ai/api/cuml/stable/](https://docs.rapids.ai/api/cuml/stable/)
- **Issues**: [https://github.com/AntonioRoye/PaCMAP-GPU/issues](https://github.com/AntonioRoye/PaCMAP-GPU/issues)

## 🙏 Acknowledgments

- Original PaCMAP authors for the algorithm and reference implementation
- RAPIDS team for GPU acceleration libraries
- PyTorch team for the deep learning framework
- NVIDIA for CUDA and GPU computing platform

---

**Made with ❤️ for the GPU computing community** 