#!/usr/bin/env python3
"""
Installation verification script for PaCMAP-GPU.
Run this script to verify that your installation is working correctly.
"""

import sys
import importlib

def check_module(module_name, required=True):
    """Check if a module can be imported."""
    try:
        importlib.import_module(module_name)
        print(f"✓ {module_name}")
        return True
    except ImportError as e:
        if required:
            print(f"✗ {module_name} - REQUIRED: {e}")
        else:
            print(f"⚠ {module_name} - OPTIONAL: {e}")
        return False

def main():
    print("PaCMAP-GPU Installation Verification")
    print("=" * 40)
    
    all_good = True
    
    # Check Python version
    print(f"\nPython Version: {sys.version}")
    if sys.version_info < (3, 10):
        print("✗ Python 3.10+ is required")
        all_good = False
    else:
        print("✓ Python version is compatible")
    
    # Check core dependencies
    print("\nChecking Core Dependencies:")
    print("-" * 30)
    
    core_modules = [
        ('numpy', True),
        ('scipy', True), 
        ('sklearn', True),
        ('matplotlib', False),
        ('pandas', False)
    ]
    
    for module, required in core_modules:
        if not check_module(module, required) and required:
            all_good = False
    
    # Check GPU dependencies
    print("\nChecking GPU Dependencies:")
    print("-" * 30)
    
    gpu_modules = [
        ('cupy', True),
        ('cuml', False),
        ('cudf', False),
    ]
    
    gpu_available = True
    for module, required in gpu_modules:
        if not check_module(module, required):
            if required:
                gpu_available = False
                all_good = False
    
    # Test GPU functionality
    if gpu_available:
        print("\nTesting GPU Functionality:")
        print("-" * 30)
        
        try:
            import cupy as cp
            
            # Test basic CuPy operations
            test_array = cp.array([1, 2, 3, 4, 5])
            result = cp.sum(test_array)
            print(f"✓ CuPy basic operations work (sum test: {result})")
            
            # Get GPU information
            try:
                device_count = cp.cuda.runtime.getDeviceCount()
                print(f"✓ Found {device_count} GPU device(s)")
                
                for i in range(device_count):
                    with cp.cuda.Device(i):
                        props = cp.cuda.runtime.getDeviceProperties(i)
                        name = props['name'].decode()
                        memory_gb = props['totalGlobalMem'] / 1e9
                        print(f"  Device {i}: {name} ({memory_gb:.1f} GB)")
                        
            except Exception as e:
                print(f"⚠ Could not get detailed GPU info: {e}")
                
        except Exception as e:
            print(f"✗ GPU functionality test failed: {e}")
            all_good = False
    
    # Test PaCMAP-GPU import
    print("\nTesting PaCMAP-GPU Import:")
    print("-" * 30)
    
    try:
        sys.path.insert(0, 'source')  # Add source directory
        
        # Import with explicit module path
        import pacmap_gpu.pacmap_gpu as pacmap_module
        import pacmap_gpu.utils as utils_module
        PaCMAPGPU = pacmap_module.PaCMAPGPU
        check_gpu_availability = utils_module.check_gpu_availability
        get_gpu_info = utils_module.get_gpu_info
        
        print("✓ PaCMAP-GPU modules imported successfully")
        
        # Debug: Check the actual class being imported
        import inspect
        sig = inspect.signature(PaCMAPGPU.__init__)
        print(f"✓ Constructor parameters: {list(sig.parameters.keys())}")
        
        if 'random_state' in sig.parameters:
            print("✓ random_state parameter is present")
        else:
            print("✗ random_state parameter is missing")
            print("Available parameters:", list(sig.parameters.keys()))
        
        # Additional debugging
        print(f"✓ PaCMAPGPU class location: {PaCMAPGPU.__module__}")
        print(f"✓ PaCMAPGPU class file: {inspect.getfile(PaCMAPGPU)}")
        
        # Test GPU availability check
        gpu_check = check_gpu_availability()
        if gpu_check:
            print("✓ GPU availability check passed")
            
            # Get detailed GPU info
            gpu_info = get_gpu_info()
            print(f"✓ GPU info retrieved: {len(gpu_info.get('devices', []))} device(s)")
            
        else:
            print("⚠ GPU not available for PaCMAP-GPU")
            
    except ImportError as e:
        print(f"✗ Could not import PaCMAP-GPU: {e}")
        all_good = False
        
    # Run basic functionality test
    if all_good and gpu_available:
        print("\nRunning Basic Functionality Test:")
        print("-" * 40)
        
        try:
            import numpy as np
            from sklearn.datasets import make_blobs
            
            # Generate test data
            X, _ = make_blobs(n_samples=100, n_features=5, centers=3, random_state=42)
            X = X.astype(np.float32)
            
            print(f"✓ Generated test dataset: {X.shape}")
            
            # Test PaCMAP-GPU
            print("Attempting to create PaCMAPGPU instance...")
            try:
                reducer = PaCMAPGPU(n_components=2, n_neighbors=5, verbose=False, random_state=42)
                print("✓ Successfully created PaCMAPGPU instance")
            except Exception as e:
                print(f"✗ Failed to create instance: {e}")
                print(f"Error type: {type(e)}")
                raise e
                
            Y = reducer.fit_transform(X)
            
            print(f"✓ PaCMAP-GPU completed successfully: {X.shape} → {Y.shape}")
            print("✓ All functionality tests passed!")
            
        except Exception as e:
            print(f"✗ Functionality test failed: {e}")
            all_good = False
    
    # Final summary
    print("\n" + "=" * 40)
    if all_good:
        print("🎉 Installation verification PASSED!")
        print("   Your PaCMAP-GPU installation is working correctly.")
        if gpu_available:
            print("   GPU acceleration is available and functional.")
        else:
            print("   Note: GPU acceleration may not be available.")
    else:
        print("❌ Installation verification FAILED!")
        print("   Please check the errors above and fix the issues.")
        print("\nCommon solutions:")
        print("- Install CUDA toolkit and drivers")
        print("- Install cupy: pip install cupy-cuda12x")
        print("- Install RAPIDS: conda install -c rapidsai -c conda-forge rapids")
        print("- Make sure you're using Python 3.10+")
    
    print("\nFor more help, see: README.md")
    return 0 if all_good else 1

if __name__ == "__main__":
    exit(main()) 