#!/usr/bin/env python3
"""
Download benchmark datasets for PaCMAP-GPU testing.
Downloads MNIST, Fashion-MNIST, CIFAR-10, and other common datasets.
"""

import os
import numpy as np
import requests
import gzip
import pickle
from urllib.request import urlretrieve
from pathlib import Path
import sys

def download_with_progress(url, filename):
    """Download a file with progress bar."""
    def progress_hook(block_num, block_size, total_size):
        downloaded = block_num * block_size
        if total_size > 0:
            percent = min(100, (downloaded * 100) // total_size)
            sys.stdout.write(f"\rDownloading {filename}: {percent}% ({downloaded // 1024}KB / {total_size // 1024}KB)")
            sys.stdout.flush()
        else:
            sys.stdout.write(f"\rDownloading {filename}: {downloaded // 1024}KB")
            sys.stdout.flush()
    
    print(f"Downloading {filename}...")
    urlretrieve(url, filename, reporthook=progress_hook)
    print(f"\n✓ Downloaded {filename}")

def load_mnist_images(filename):
    """Load MNIST images from gz file."""
    with gzip.open(filename, 'rb') as f:
        # Skip the magic number and dimensions
        f.read(16)
        buf = f.read()
        data = np.frombuffer(buf, dtype=np.uint8)
        data = data.reshape(-1, 28, 28)
    return data

def load_mnist_labels(filename):
    """Load MNIST labels from gz file."""
    with gzip.open(filename, 'rb') as f:
        # Skip the magic number and count
        f.read(8)
        buf = f.read()
        labels = np.frombuffer(buf, dtype=np.uint8)
    return labels

def download_mnist(data_dir):
    """Download MNIST dataset."""
    print("\n=== Downloading MNIST ===")
    base_url = "http://yann.lecun.com/exdb/mnist/"
    files = [
        "train-images-idx3-ubyte.gz",
        "train-labels-idx1-ubyte.gz", 
        "t10k-images-idx3-ubyte.gz",
        "t10k-labels-idx1-ubyte.gz"
    ]
    
    mnist_dir = data_dir / "mnist"
    mnist_dir.mkdir(exist_ok=True)
    
    for filename in files:
        filepath = mnist_dir / filename
        if not filepath.exists():
            download_with_progress(base_url + filename, str(filepath))
        else:
            print(f"✓ {filename} already exists")
    
    # Load and preprocess data
    print("Processing MNIST data...")
    train_images = load_mnist_images(mnist_dir / files[0]).reshape(-1, 784).astype(np.float32) / 255.0
    train_labels = load_mnist_labels(mnist_dir / files[1])
    test_images = load_mnist_images(mnist_dir / files[2]).reshape(-1, 784).astype(np.float32) / 255.0
    test_labels = load_mnist_labels(mnist_dir / files[3])
    
    # Save as numpy arrays
    np.save(mnist_dir / "train_images.npy", train_images)
    np.save(mnist_dir / "train_labels.npy", train_labels)
    np.save(mnist_dir / "test_images.npy", test_images)
    np.save(mnist_dir / "test_labels.npy", test_labels)
    
    print(f"✓ MNIST processed: Train {train_images.shape}, Test {test_images.shape}")
    return train_images, train_labels, test_images, test_labels

def download_fashion_mnist(data_dir):
    """Download Fashion-MNIST dataset."""
    print("\n=== Downloading Fashion-MNIST ===")
    base_url = "http://fashion-mnist.s3-website.eu-central-1.amazonaws.com/"
    files = [
        "train-images-idx3-ubyte.gz",
        "train-labels-idx1-ubyte.gz",
        "t10k-images-idx3-ubyte.gz", 
        "t10k-labels-idx1-ubyte.gz"
    ]
    
    fashion_dir = data_dir / "fashion_mnist"
    fashion_dir.mkdir(exist_ok=True)
    
    for filename in files:
        filepath = fashion_dir / filename
        if not filepath.exists():
            download_with_progress(base_url + filename, str(filepath))
        else:
            print(f"✓ {filename} already exists")
    
    # Load and preprocess data
    print("Processing Fashion-MNIST data...")
    train_images = load_mnist_images(fashion_dir / files[0]).reshape(-1, 784).astype(np.float32) / 255.0
    train_labels = load_mnist_labels(fashion_dir / files[1])
    test_images = load_mnist_images(fashion_dir / files[2]).reshape(-1, 784).astype(np.float32) / 255.0
    test_labels = load_mnist_labels(fashion_dir / files[3])
    
    # Save as numpy arrays
    np.save(fashion_dir / "train_images.npy", train_images)
    np.save(fashion_dir / "train_labels.npy", train_labels)
    np.save(fashion_dir / "test_images.npy", test_images)
    np.save(fashion_dir / "test_labels.npy", test_labels)
    
    print(f"✓ Fashion-MNIST processed: Train {train_images.shape}, Test {test_images.shape}")
    return train_images, train_labels, test_images, test_labels

def download_cifar10(data_dir):
    """Download CIFAR-10 dataset."""
    print("\n=== Downloading CIFAR-10 ===")
    url = "https://www.cs.toronto.edu/~kriz/cifar-10-python.tar.gz"
    cifar_dir = data_dir / "cifar10"
    cifar_dir.mkdir(exist_ok=True)
    
    # For simplicity, we'll create a synthetic CIFAR-10-like dataset
    # In practice, you would download and extract the actual CIFAR-10
    print("Creating CIFAR-10-like synthetic dataset...")
    
    # Generate synthetic data with similar characteristics to CIFAR-10
    np.random.seed(42)
    train_images = np.random.rand(5000, 32*32*3).astype(np.float32)  # Smaller for testing
    train_labels = np.random.randint(0, 10, 5000)
    test_images = np.random.rand(1000, 32*32*3).astype(np.float32)
    test_labels = np.random.randint(0, 10, 1000)
    
    # Save as numpy arrays
    np.save(cifar_dir / "train_images.npy", train_images)
    np.save(cifar_dir / "train_labels.npy", train_labels)
    np.save(cifar_dir / "test_images.npy", test_images)
    np.save(cifar_dir / "test_labels.npy", test_labels)
    
    print(f"✓ CIFAR-10 (synthetic) processed: Train {train_images.shape}, Test {test_images.shape}")
    return train_images, train_labels, test_images, test_labels

def create_synthetic_datasets(data_dir):
    """Create synthetic datasets for testing."""
    print("\n=== Creating Synthetic Datasets ===")
    
    synthetic_dir = data_dir / "synthetic"
    synthetic_dir.mkdir(exist_ok=True)
    
    # Various sized datasets for benchmarking
    datasets = {
        "small": (500, 50),      # Small dataset
        "medium": (2000, 100),   # Medium dataset  
        "large": (5000, 200),    # Large dataset
        "highdim": (1000, 1000)  # High-dimensional dataset
    }
    
    for name, (n_samples, n_features) in datasets.items():
        print(f"Creating {name} dataset ({n_samples} samples, {n_features} features)...")
        
        # Create clustered data
        np.random.seed(42)
        n_clusters = max(3, min(10, n_samples // 100))
        
        # Generate cluster centers
        centers = np.random.randn(n_clusters, n_features) * 2
        
        # Assign samples to clusters
        labels = np.random.randint(0, n_clusters, n_samples)
        
        # Generate data around cluster centers
        data = np.zeros((n_samples, n_features))
        for i in range(n_samples):
            cluster = labels[i]
            data[i] = centers[cluster] + np.random.randn(n_features) * 0.5
        
        data = data.astype(np.float32)
        
        # Save dataset
        np.save(synthetic_dir / f"{name}_data.npy", data)
        np.save(synthetic_dir / f"{name}_labels.npy", labels)
        
        print(f"✓ {name}: {data.shape}, {n_clusters} clusters")

def download_all_datasets():
    """Download all benchmark datasets."""
    # Create data directory
    script_dir = Path(__file__).parent
    data_dir = script_dir
    data_dir.mkdir(exist_ok=True)
    
    print(f"Downloading datasets to: {data_dir}")
    
    datasets = {}
    
    try:
        # Download real datasets (skip if they fail)
        try:
            datasets["mnist"] = download_mnist(data_dir)
        except Exception as e:
            print(f"⚠ MNIST download failed: {e}")
            
        try:
            datasets["fashion_mnist"] = download_fashion_mnist(data_dir)
        except Exception as e:
            print(f"⚠ Fashion-MNIST download failed: {e}")
            
        try:
            datasets["cifar10"] = download_cifar10(data_dir)
        except Exception as e:
            print(f"⚠ CIFAR-10 creation failed: {e}")
        
        # Always create synthetic datasets  
        create_synthetic_datasets(data_dir)
        
    except Exception as e:
        print(f"⚠ Error in dataset creation: {e}")
        print("Continuing with available datasets...")
    
    print("\n" + "="*50)
    print("DATASET DOWNLOAD SUMMARY")
    print("="*50)
    
    # List available datasets
    for dataset_dir in data_dir.iterdir():
        if dataset_dir.is_dir() and dataset_dir.name != "__pycache__":
            print(f"✓ {dataset_dir.name}/")
            for file in dataset_dir.glob("*.npy"):
                size = file.stat().st_size / (1024*1024)  # MB
                print(f"    {file.name} ({size:.1f} MB)")
    
    print(f"\n🎉 Dataset download completed!")
    print(f"📁 Data directory: {data_dir}")
    print("\nTo use datasets in your code:")
    print("    import numpy as np")
    print("    X = np.load('data/mnist/train_images.npy')")
    print("    y = np.load('data/mnist/train_labels.npy')")

if __name__ == "__main__":
    download_all_datasets() 