#!/usr/bin/env python3
"""
Basic demo of PaCMAP-GPU functionality.

This script demonstrates the basic usage of PaCMAP-GPU on a simple dataset.
"""

import numpy as np
import matplotlib.pyplot as plt
from sklearn.datasets import make_swiss_roll, load_digits
import sys
import os

# Add source directory to path
sys.path.insert(0, os.path.join(os.path.dirname(__file__), '..', 'source'))

from pacmap_gpu import PaCMAPGPU, check_gpu_availability, print_installation_report


def demo_swiss_roll():
    """Demonstrate PaCMAP-GPU on Swiss roll dataset."""
    print("\n" + "="*60)
    print("DEMO 1: Swiss Roll Dataset")
    print("="*60)
    
    # Generate Swiss roll dataset
    print("Generating Swiss roll dataset...")
    X, color = make_swiss_roll(n_samples=1000, noise=0.1, random_state=42)
    X = X.astype(np.float32)
    
    print(f"Dataset shape: {X.shape}")
    
    # Apply PaCMAP-GPU
    print("\nRunning PaCMAP-GPU...")
    reducer = PaCMAPGPU(
        n_components=2,
        n_neighbors=10,
        num_iters=200,  # Reduced for demo
        verbose=True,
        random_state=42
    )
    
    X_embedded = reducer.fit_transform(X)
    
    print(f"Embedding shape: {X_embedded.shape}")
    
    # Plot results
    plt.figure(figsize=(12, 5))
    
    # Original 3D data
    plt.subplot(1, 2, 1)
    plt.scatter(X[:, 0], X[:, 1], c=color, cmap='viridis', alpha=0.7)
    plt.title('Original Swiss Roll (X-Y view)')
    plt.xlabel('X')
    plt.ylabel('Y')
    plt.colorbar()
    
    # 2D embedding
    plt.subplot(1, 2, 2)
    plt.scatter(X_embedded[:, 0], X_embedded[:, 1], c=color, cmap='viridis', alpha=0.7)
    plt.title('PaCMAP-GPU Embedding')
    plt.xlabel('Component 1')
    plt.ylabel('Component 2')
    plt.colorbar()
    
    plt.tight_layout()
    plt.savefig('swiss_roll_demo.png', dpi=150, bbox_inches='tight')
    print("\nPlot saved as 'swiss_roll_demo.png'")
    
    return X_embedded


def demo_digits():
    """Demonstrate PaCMAP-GPU on digits dataset."""
    print("\n" + "="*60)
    print("DEMO 2: Digits Dataset")
    print("="*60)
    
    # Load digits dataset
    print("Loading digits dataset...")
    digits = load_digits()
    X, y = digits.data, digits.target
    X = X.astype(np.float32)
    
    print(f"Dataset shape: {X.shape}")
    print(f"Number of classes: {len(np.unique(y))}")
    
    # Apply PaCMAP-GPU
    print("\nRunning PaCMAP-GPU...")
    reducer = PaCMAPGPU(
        n_components=2,
        n_neighbors=15,
        num_iters=300,  # More iterations for better results
        verbose=True,
        random_state=42
    )
    
    X_embedded = reducer.fit_transform(X)
    
    print(f"Embedding shape: {X_embedded.shape}")
    
    # Plot results
    plt.figure(figsize=(12, 5))
    
    # Show some original digit images
    plt.subplot(1, 2, 1)
    for i in range(10):
        plt.subplot(2, 10, i + 1)
        plt.imshow(digits.images[i], cmap='gray')
        plt.title(f'{digits.target[i]}')
        plt.axis('off')
    
    # 2D embedding colored by digit class
    plt.subplot(1, 2, 2)
    scatter = plt.scatter(X_embedded[:, 0], X_embedded[:, 1], 
                         c=y, cmap='tab10', alpha=0.7, s=20)
    plt.title('PaCMAP-GPU Embedding (colored by digit)')
    plt.xlabel('Component 1')
    plt.ylabel('Component 2')
    
    # Add legend
    for i in range(10):
        plt.scatter([], [], c=plt.cm.tab10(i/9), label=f'Digit {i}')
    plt.legend(bbox_to_anchor=(1.05, 1), loc='upper left')
    
    plt.tight_layout()
    plt.savefig('digits_demo.png', dpi=150, bbox_inches='tight')
    print("\nPlot saved as 'digits_demo.png'")
    
    return X_embedded


def demo_performance_comparison():
    """Compare performance with different parameters."""
    print("\n" + "="*60)
    print("DEMO 3: Performance Comparison")
    print("="*60)
    
    # Generate test dataset
    from sklearn.datasets import make_blobs
    X, _ = make_blobs(n_samples=2000, n_features=50, centers=10, random_state=42)
    X = X.astype(np.float32)
    
    print(f"Dataset shape: {X.shape}")
    
    # Test different configurations
    configs = [
        {"name": "Fast", "num_iters": 100, "batch_size": 16384},
        {"name": "Balanced", "num_iters": 300, "batch_size": 8192},
        {"name": "High Quality", "num_iters": 500, "batch_size": 4096},
    ]
    
    results = []
    
    for config in configs:
        print(f"\nTesting {config['name']} configuration...")
        
        reducer = PaCMAPGPU(
            n_components=2,
            n_neighbors=10,
            num_iters=config['num_iters'],
            batch_size=config['batch_size'],
            verbose=False,
            random_state=42
        )
        
        import time
        start_time = time.perf_counter()
        X_embedded = reducer.fit_transform(X)
        elapsed_time = time.perf_counter() - start_time
        
        results.append({
            'config': config['name'],
            'time': elapsed_time,
            'embedding': X_embedded
        })
        
        print(f"  Time: {elapsed_time:.2f}s")
        print(f"  Embedding std: [{np.std(X_embedded[:, 0]):.3f}, {np.std(X_embedded[:, 1]):.3f}]")
    
    # Plot comparison
    plt.figure(figsize=(15, 5))
    
    for i, result in enumerate(results):
        plt.subplot(1, 3, i + 1)
        plt.scatter(result['embedding'][:, 0], result['embedding'][:, 1], 
                   alpha=0.6, s=10)
        plt.title(f"{result['config']}\n({result['time']:.2f}s)")
        plt.xlabel('Component 1')
        plt.ylabel('Component 2')
    
    plt.suptitle('Performance vs Quality Trade-off')
    plt.tight_layout()
    plt.savefig('performance_comparison.png', dpi=150, bbox_inches='tight')
    print("\nComparison plot saved as 'performance_comparison.png'")
    
    return results


def main():
    """Main demo function."""
    print("PaCMAP-GPU Basic Demo")
    print("="*60)
    
    # Check installation
    print("\nChecking installation...")
    print_installation_report()
    
    # Check GPU availability
    if not check_gpu_availability():
        print("\n⚠️ WARNING: GPU not available. This demo requires CUDA.")
        print("Please install CUDA and GPU dependencies for best performance.")
        return
    
    print("\n✅ GPU is available! Running demos...")
    
    try:
        # Run demos
        demo_swiss_roll()
        demo_digits()
        demo_performance_comparison()
        
        print("\n" + "="*60)
        print("🎉 Demo completed successfully!")
        print("📁 Check the generated PNG files for visualizations.")
        print("="*60)
        
    except Exception as e:
        print(f"\n❌ Demo failed with error: {e}")
        import traceback
        traceback.print_exc()


if __name__ == "__main__":
    main() 