#!/usr/bin/env python3
"""
Setup script for PaCMAP-GPU: GPU-Accelerated Pairwise Controlled Manifold Approximation
"""

import os
import sys
from setuptools import setup, find_packages

# Ensure Python 3.10+
if sys.version_info < (3, 10):
    sys.exit("Python 3.10 or higher is required. You have Python {}.{}.".format(*sys.version_info[:2]))

def read_requirements(filename):
    """Read requirements from file."""
    filepath = os.path.join(os.path.dirname(__file__), filename)
    if os.path.exists(filepath):
        with open(filepath, 'r', encoding='utf-8') as f:
            return [line.strip() for line in f.readlines() 
                   if line.strip() and not line.startswith('#') and not line.startswith('-')]
    return []

def get_version():
    """Get version from __init__.py file."""
    version_file = os.path.join(os.path.dirname(__file__), 'source', 'pacmap_gpu', '__init__.py')
    with open(version_file, 'r', encoding='utf-8') as f:
        for line in f:
            if line.startswith('__version__'):
                return line.split('=')[1].strip().strip('"').strip("'")
    return "0.1.0"

def get_long_description():
    """Get long description from README."""
    readme_path = os.path.join(os.path.dirname(__file__), 'README.md')
    if os.path.exists(readme_path):
        with open(readme_path, 'r', encoding='utf-8') as f:
            return f.read()
    return ""

# Core dependencies that are always required
install_requires = [
    'torch>=2.0.0',
    'numpy>=1.21.0,<2.0.0',  # <2 for FAISS compatibility
    'scipy>=1.7.0',
    'scikit-learn>=1.0.0',
    'pandas>=1.3.0',
    'matplotlib>=3.5.0',
    # GPU dependencies - required since code imports cuML directly
    'cupy-cuda12x>=12.0.0',
    'cuml-cu12>=24.0.0',
    'cudf-cu12>=24.0.0',
    'faiss-gpu>=1.7.0',
]

# Additional dependencies for development
dev_requires = [
    'pytest>=6.0.0',
    'pytest-benchmark>=4.0.0',
    'pytest-cov>=4.0.0',
    'black>=22.0.0',
    'flake8>=4.0.0',
    'mypy>=0.950',
    'pre-commit>=2.17.0',
    'jupyterlab>=3.0.0',
]

# Documentation dependencies
docs_requires = [
    'sphinx>=4.0.0',
    'sphinx-rtd-theme>=1.0.0',
    'myst-parser>=0.17.0',
    'nbsphinx>=0.8.0',
]

setup(
    name="pacmap-gpu",
    version=get_version(),
    author="Antonio Roye-Azar",
    author_email="antonio@varonova.ca",
    description="GPU-Accelerated Pairwise Controlled Manifold Approximation for large-scale dimensionality reduction",
    long_description=get_long_description(),
    long_description_content_type="text/markdown",
    url="https://github.com/AntonioRoye/PaCMAP-GPU",
    project_urls={
        "Bug Reports": "https://github.com/AntonioRoye/PaCMAP-GPU/issues",
        "Source": "https://github.com/AntonioRoye/PaCMAP-GPU",
        "Documentation": "https://github.com/AntonioRoye/PaCMAP-GPU#readme",
    },
    packages=find_packages(where="source"),
    package_dir={"": "source"},
    package_data={
        "pacmap_gpu": ["*.py"],
    },
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Science/Research",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: POSIX :: Linux",
        "Operating System :: Microsoft :: Windows",
        "Operating System :: MacOS",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
        "Topic :: Scientific/Engineering :: Information Analysis",
        "Topic :: Scientific/Engineering :: Visualization",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
    keywords=[
        "dimensionality-reduction", "manifold-learning", "gpu", "cuda", 
        "machine-learning", "data-visualization", "pytorch", "rapids"
    ],
    python_requires=">=3.10",
    install_requires=install_requires,
    extras_require={
        "dev": dev_requires,
        "docs": docs_requires,
        "all": dev_requires + docs_requires,
    },
    entry_points={
        "console_scripts": [
            "pacmap-gpu-check=pacmap_gpu.utils:main_check_installation",
        ],
    },
    zip_safe=False,
    include_package_data=True,
    license="Apache-2.0",
    platforms=["Linux", "Windows", "macOS"],
) 