"""
Pytest configuration and shared fixtures for PaCMAP-GPU tests.
"""

import pytest
import numpy as np
import sys
import os
from typing import Generator, Tuple
from sklearn.datasets import make_blobs, make_swiss_roll

# Add source directory to path for imports
sys.path.insert(0, os.path.join(os.path.dirname(__file__), '..', 'source'))

try:
    import torch
    TORCH_AVAILABLE = torch.cuda.is_available()
except ImportError:
    TORCH_AVAILABLE = False

try:
    from pacmap_gpu import PaCMAPGPU, check_gpu_availability
    PACMAP_GPU_AVAILABLE = True
except ImportError as e:
    PACMAP_GPU_AVAILABLE = False
    print(f"PaCMAP-GPU import failed: {e}")
    # Create dummy functions for when imports fail
    def check_gpu_availability():
        return False


def pytest_configure(config):
    """Configure pytest with custom markers."""
    config.addinivalue_line("markers", "gpu: mark test as requiring GPU")
    config.addinivalue_line("markers", "slow: mark test as slow running")
    config.addinivalue_line("markers", "benchmark: mark test as benchmark")


def pytest_collection_modifyitems(config, items):
    """Modify test collection to skip GPU tests if GPU is not available."""
    gpu_available = TORCH_AVAILABLE and PACMAP_GPU_AVAILABLE
    if gpu_available:
        try:
            gpu_available = check_gpu_availability()
        except Exception:
            gpu_available = False
    
    if not gpu_available:
        skip_gpu = pytest.mark.skip(reason="GPU not available or PaCMAP-GPU not properly installed")
        for item in items:
            if "gpu" in item.keywords:
                item.add_marker(skip_gpu)


@pytest.fixture
def small_2d_data() -> np.ndarray:
    """Generate small 2D dataset for quick tests."""
    np.random.seed(42)
    return np.random.randn(50, 5).astype(np.float32)


@pytest.fixture  
def medium_data() -> np.ndarray:
    """Generate medium-sized dataset for standard tests."""
    X, _ = make_blobs(n_samples=200, n_features=10, centers=5, random_state=42)
    return X.astype(np.float32)


@pytest.fixture
def large_data() -> np.ndarray:
    """Generate large dataset for performance tests."""
    X, _ = make_blobs(n_samples=1000, n_features=50, centers=10, random_state=42)
    return X.astype(np.float32)


@pytest.fixture
def swiss_roll_data() -> Tuple[np.ndarray, np.ndarray]:
    """Generate Swiss roll manifold dataset."""
    X, color = make_swiss_roll(n_samples=500, noise=0.1, random_state=42)
    return X.astype(np.float32), color


@pytest.fixture
def high_dimensional_data() -> np.ndarray:
    """Generate high-dimensional dataset to test PCA preprocessing."""
    X, _ = make_blobs(n_samples=100, n_features=200, centers=5, random_state=42)
    return X.astype(np.float32)


@pytest.fixture
def pacmap_gpu_instance():
    """Create a PaCMAPGPU instance with default parameters."""
    if not PACMAP_GPU_AVAILABLE:
        pytest.skip("PaCMAP-GPU not available")
    
    try:
        reducer = PaCMAPGPU(n_components=2, n_neighbors=5, num_iters=10, verbose=False, random_state=42)
        yield reducer
        
        # Cleanup
        try:
            from pacmap_gpu.utils import clear_gpu_memory
            clear_gpu_memory()
        except Exception:
            pass
    except Exception as e:
        pytest.skip(f"Could not create PaCMAPGPU instance: {e}")


@pytest.fixture
def sample_datasets():
    """Load sample datasets from data directory if available."""
    data_dir = os.path.join(os.path.dirname(__file__), '..', 'data')
    datasets = {}
    
    # Try to load existing datasets
    dataset_paths = {
        'mnist': os.path.join(data_dir, 'mnist', 'test_images.npy'),
        'fashion_mnist': os.path.join(data_dir, 'fashion_mnist', 'test_images.npy'),
        'synthetic_small': os.path.join(data_dir, 'synthetic', 'small_data.npy'),
        'synthetic_medium': os.path.join(data_dir, 'synthetic', 'medium_data.npy')
    }
    
    for name, path in dataset_paths.items():
        if os.path.exists(path):
            try:
                datasets[name] = np.load(path).astype(np.float32)
                # Subsample for faster testing
                if datasets[name].shape[0] > 1000:
                    datasets[name] = datasets[name][:1000]
            except Exception:
                continue
    
    return datasets


@pytest.fixture(scope="session", autouse=True)
def setup_test_environment():
    """Set up the test environment."""
    # Ensure reproducible results
    np.random.seed(42)
    
    if TORCH_AVAILABLE:
        torch.manual_seed(42)
        if torch.cuda.is_available():
            torch.cuda.manual_seed(42)
    
    yield
    
    # Cleanup after all tests
    if TORCH_AVAILABLE and torch.cuda.is_available():
        try:
            torch.cuda.empty_cache()
            torch.cuda.synchronize()
        except Exception:
            pass 