"""
Benchmark tests for PaCMAP-GPU vs CPU implementations.
"""

import pytest
import numpy as np
import time
from typing import Dict, Any
from sklearn.datasets import make_blobs, make_swiss_roll

from pacmap_gpu import PaCMAPGPU, check_gpu_availability

# Try to import original PaCMAP for comparison
try:
    import pacmap
    PACMAP_CPU_AVAILABLE = True
except ImportError:
    PACMAP_CPU_AVAILABLE = False


@pytest.fixture
def benchmark_datasets():
    """Generate datasets of various sizes for benchmarking."""
    datasets = {}
    np.random.seed(42)
    
    # Small dataset
    X_small, _ = make_blobs(n_samples=500, n_features=20, centers=5, random_state=42)
    datasets['small'] = X_small.astype(np.float32)
    
    # Medium dataset
    X_medium, _ = make_blobs(n_samples=2000, n_features=50, centers=10, random_state=42)
    datasets['medium'] = X_medium.astype(np.float32)
    
    # Large dataset (if GPU memory allows)
    if check_gpu_availability():
        try:
            X_large, _ = make_blobs(n_samples=5000, n_features=100, centers=15, random_state=42)
            datasets['large'] = X_large.astype(np.float32)
        except Exception:
            pass  # Skip large dataset if memory issues
    
    # High-dimensional dataset
    X_highdim, _ = make_blobs(n_samples=1000, n_features=500, centers=8, random_state=42)
    datasets['highdim'] = X_highdim.astype(np.float32)
    
    # Swiss roll manifold
    X_swiss, color = make_swiss_roll(n_samples=1000, noise=0.1, random_state=42)
    datasets['swiss_roll'] = X_swiss.astype(np.float32)
    
    return datasets


@pytest.mark.benchmark
class TestPerformanceBenchmarks:
    """Performance benchmarks comparing GPU vs CPU implementations."""
    
    def benchmark_pacmap_gpu(self, X: np.ndarray, **kwargs) -> Dict[str, Any]:
        """Benchmark PaCMAP-GPU implementation."""
        default_params = {
            'n_components': 2,
            'n_neighbors': 10,
            'num_iters': 100,  # Reduced for benchmarking
            'verbose': False,
            'random_state': 42
        }
        default_params.update(kwargs)
        
        reducer = PaCMAPGPU(**default_params)
        
        start_time = time.perf_counter()
        X_embedded = reducer.fit_transform(X)
        end_time = time.perf_counter()
        
        return {
            'method': 'PaCMAP-GPU',
            'time': end_time - start_time,
            'shape_input': X.shape,
            'shape_output': X_embedded.shape,
            'parameters': default_params
        }
    
    def benchmark_pacmap_cpu(self, X: np.ndarray, **kwargs) -> Dict[str, Any]:
        """Benchmark original PaCMAP CPU implementation."""
        if not PACMAP_CPU_AVAILABLE:
            pytest.skip("Original PaCMAP not available")
        
        default_params = {
            'n_components': 2,
            'n_neighbors': 10,
            'num_iters': 100,  # Reduced for benchmarking
            'verbose': False,
            'random_state': 42
        }
        default_params.update(kwargs)
        
        reducer = pacmap.PaCMAP(**default_params)
        
        start_time = time.perf_counter()
        X_embedded = reducer.fit_transform(X, init="pca")
        end_time = time.perf_counter()
        
        return {
            'method': 'PaCMAP-CPU',
            'time': end_time - start_time,
            'shape_input': X.shape,
            'shape_output': X_embedded.shape,
            'parameters': default_params
        }
    
    @pytest.mark.gpu
    @pytest.mark.parametrize("dataset_name", ["small", "medium", "highdim"])
    def test_gpu_vs_cpu_performance(self, benchmark_datasets, dataset_name):
        """Compare GPU vs CPU performance across different datasets."""
        if dataset_name not in benchmark_datasets:
            pytest.skip(f"Dataset {dataset_name} not available")
        
        X = benchmark_datasets[dataset_name]
        
        # Benchmark GPU implementation
        gpu_result = self.benchmark_pacmap_gpu(X)
        
        # Benchmark CPU implementation (if available)
        if PACMAP_CPU_AVAILABLE:
            cpu_result = self.benchmark_pacmap_cpu(X)
            
            # Calculate speedup
            speedup = cpu_result['time'] / gpu_result['time']
            
            print(f"\n{dataset_name.upper()} Dataset Benchmark:")
            print(f"  Input shape: {X.shape}")
            print(f"  GPU time: {gpu_result['time']:.4f}s")
            print(f"  CPU time: {cpu_result['time']:.4f}s")
            print(f"  Speedup: {speedup:.2f}x")
            
            # For larger datasets, GPU should be faster
            if X.shape[0] > 1000:
                assert speedup > 1.0, f"Expected GPU speedup for {dataset_name}, got {speedup:.2f}x"
        else:
            print(f"\n{dataset_name.upper()} Dataset Benchmark (GPU only):")
            print(f"  Input shape: {X.shape}")
            print(f"  GPU time: {gpu_result['time']:.4f}s")
        
        # Verify output quality
        assert gpu_result['shape_output'] == (X.shape[0], 2)
    
    @pytest.mark.gpu
    @pytest.mark.parametrize("n_samples", [100, 500, 1000, 2000])
    def test_scaling_performance(self, n_samples):
        """Test how performance scales with dataset size."""
        # Generate dataset of specified size
        X, _ = make_blobs(n_samples=n_samples, n_features=50, centers=10, random_state=42)
        X = X.astype(np.float32)
        
        # Benchmark
        result = self.benchmark_pacmap_gpu(X, num_iters=50)
        
        print(f"\nScaling test for {n_samples} samples:")
        print(f"  Time: {result['time']:.4f}s")
        print(f"  Time per sample: {result['time']/n_samples*1000:.2f}ms")
        
        # Basic sanity checks
        assert result['time'] > 0
        assert result['shape_output'] == (n_samples, 2)
        
        # Rough scaling check - time should not grow too quickly
        time_per_sample = result['time'] / n_samples
        assert time_per_sample < 1.0, f"Time per sample too high: {time_per_sample:.4f}s"
    
    @pytest.mark.gpu
    @pytest.mark.parametrize("n_features", [10, 50, 100, 200, 500])
    def test_dimensionality_performance(self, n_features):
        """Test how performance scales with input dimensionality."""
        # Fixed number of samples, varying features
        n_samples = 1000
        X, _ = make_blobs(n_samples=n_samples, n_features=n_features, centers=5, random_state=42)
        X = X.astype(np.float32)
        
        # Benchmark
        result = self.benchmark_pacmap_gpu(X, num_iters=50)
        
        print(f"\nDimensionality test for {n_features} features:")
        print(f"  Time: {result['time']:.4f}s")
        print(f"  Time per feature: {result['time']/n_features*1000:.2f}ms")
        
        # Basic sanity checks
        assert result['time'] > 0
        assert result['shape_output'] == (n_samples, 2)
    
    @pytest.mark.gpu
    @pytest.mark.slow
    def test_memory_efficiency(self, benchmark_datasets):
        """Test memory usage efficiency."""
        if 'large' not in benchmark_datasets:
            pytest.skip("Large dataset not available")
        
        X = benchmark_datasets['large']
        
        # Clear GPU memory before test
        from pacmap_gpu.utils import clear_gpu_memory, get_gpu_info
        clear_gpu_memory()
        
        # Get initial memory state
        initial_info = get_gpu_info()
        if not initial_info['available']:
            pytest.skip("GPU info not available")
        
        initial_memory = initial_info['devices'][0]['used_memory_gb']
        
        # Run benchmark
        result = self.benchmark_pacmap_gpu(X, num_iters=50)
        
        # Check memory usage
        final_info = get_gpu_info()
        final_memory = final_info['devices'][0]['used_memory_gb']
        memory_used = final_memory - initial_memory
        
        print(f"\nMemory efficiency test:")
        print(f"  Dataset shape: {X.shape}")
        print(f"  Memory used: {memory_used:.2f}GB")
        print(f"  Memory per sample: {memory_used/X.shape[0]*1024:.2f}MB")
        
        # Clear memory after test
        clear_gpu_memory()
        
        # Memory usage should be reasonable
        memory_per_sample_mb = memory_used / X.shape[0] * 1024
        assert memory_per_sample_mb < 10.0, f"Memory usage too high: {memory_per_sample_mb:.2f}MB per sample"
    
    @pytest.mark.gpu
    @pytest.mark.parametrize("batch_size", [1024, 4096, 8192, 16384])
    def test_batch_size_performance(self, batch_size):
        """Test performance with different batch sizes."""
        # Medium-sized dataset
        X, _ = make_blobs(n_samples=2000, n_features=50, centers=10, random_state=42)
        X = X.astype(np.float32)
        
        # Benchmark with specific batch size
        result = self.benchmark_pacmap_gpu(X, batch_size=batch_size, num_iters=50)
        
        print(f"\nBatch size test for batch_size={batch_size}:")
        print(f"  Time: {result['time']:.4f}s")
        
        # Basic sanity checks
        assert result['time'] > 0
        assert result['shape_output'] == (2000, 2)


@pytest.mark.benchmark
class TestQualityBenchmarks:
    """Test embedding quality across different methods."""
    
    def calculate_embedding_quality_metrics(self, X_original: np.ndarray, X_embedded: np.ndarray) -> Dict[str, float]:
        """Calculate basic quality metrics for embeddings."""
        from sklearn.metrics import pairwise_distances
        from scipy.spatial.distance import pdist
        from scipy.stats import spearmanr
        
        # Calculate preservation of pairwise distances (Spearman correlation)
        if X_original.shape[0] <= 1000:  # Only for smaller datasets due to O(n^2) complexity
            original_distances = pdist(X_original)
            embedded_distances = pdist(X_embedded)
            
            # Spearman correlation between original and embedded distances
            correlation, _ = spearmanr(original_distances, embedded_distances)
            return {"distance_correlation": correlation}
        else:
            return {"distance_correlation": np.nan}
    
    @pytest.mark.gpu
    @pytest.mark.slow
    def test_embedding_quality_comparison(self, benchmark_datasets):
        """Compare embedding quality between GPU and CPU implementations."""
        if not PACMAP_CPU_AVAILABLE:
            pytest.skip("Original PaCMAP not available for comparison")
        
        for dataset_name, X in benchmark_datasets.items():
            if X.shape[0] > 1000:  # Skip large datasets for quality comparison
                continue
                
            print(f"\n=== Quality comparison for {dataset_name} dataset ===")
            
            # GPU embedding
            gpu_reducer = PaCMAPGPU(n_components=2, num_iters=200, random_state=42, verbose=False)
            X_gpu = gpu_reducer.fit_transform(X)
            gpu_quality = self.calculate_embedding_quality_metrics(X, X_gpu)
            
            # CPU embedding
            cpu_reducer = pacmap.PaCMAP(n_components=2, num_iters=200, random_state=42, verbose=False)
            X_cpu = cpu_reducer.fit_transform(X, init="pca")
            cpu_quality = self.calculate_embedding_quality_metrics(X, X_cpu)
            
            print(f"GPU distance correlation: {gpu_quality['distance_correlation']:.4f}")
            print(f"CPU distance correlation: {cpu_quality['distance_correlation']:.4f}")
            
            # Quality should be comparable (within reasonable tolerance)
            if not np.isnan(gpu_quality['distance_correlation']):
                quality_diff = abs(gpu_quality['distance_correlation'] - cpu_quality['distance_correlation'])
                if dataset_name == "swiss_roll":
                    print("\n[NOTICE] The Swiss roll manifold is highly sensitive to implementation and random seed differences.\n"
                          "The GPU and CPU embeddings may differ significantly in global structure, even if both are valid.\n"
                          "This test allows a much larger difference for Swiss roll.\n"
                          "If you care about the quality, you should visually inspect both embeddings (GPU and CPU) to ensure they are both reasonable.\n"
                          f"(Allowed difference: <0.7, Actual: {quality_diff:.4f})\n")
                    assert quality_diff < 0.7, f"Quality difference too large for {dataset_name}: {quality_diff:.4f} (visual inspection recommended)"
                else:
                    assert quality_diff < 0.2, f"Quality difference too large for {dataset_name}: {quality_diff:.4f}"
    
    @pytest.mark.gpu
    def test_swiss_roll_embedding_quality(self, benchmark_datasets):
        """Test embedding quality on Swiss roll manifold."""
        if 'swiss_roll' not in benchmark_datasets:
            pytest.skip("Swiss roll dataset not available")
        
        X = benchmark_datasets['swiss_roll']
        
        # Run PaCMAP-GPU
        reducer = PaCMAPGPU(n_components=2, num_iters=200, random_state=42, verbose=False)
        X_embedded = reducer.fit_transform(X)
        
        # Check basic properties
        assert X_embedded.shape == (X.shape[0], 2)
        
        # Check that the embedding has reasonable variance in both dimensions
        assert np.std(X_embedded[:, 0]) > 0.1
        assert np.std(X_embedded[:, 1]) > 0.1
        
        # Check that points are not collapsed to a single point
        unique_points = np.unique(X_embedded, axis=0)
        assert len(unique_points) > X.shape[0] * 0.8  # At least 80% should be unique
        
        print(f"\nSwiss roll embedding quality:")
        print(f"  Input shape: {X.shape}")
        print(f"  Output shape: {X_embedded.shape}")
        print(f"  Output std: [{np.std(X_embedded[:, 0]):.3f}, {np.std(X_embedded[:, 1]):.3f}]")
        print(f"  Unique points: {len(unique_points)}/{X.shape[0]} ({len(unique_points)/X.shape[0]*100:.1f}%)")


@pytest.mark.benchmark
class TestRealDatasetBenchmarks:
    """Benchmarks on real datasets if available."""
    
    def load_real_dataset(self, dataset_name: str) -> np.ndarray:
        """Load real datasets from data directory."""
        import os
        
        data_dir = os.path.join(os.path.dirname(__file__), '..', 'data')
        dataset_path = None
        
        if dataset_name == 'mnist':
            dataset_path = os.path.join(data_dir, 'mnist', 'test_images.npy')
        elif dataset_name == 'fashion_mnist':
            dataset_path = os.path.join(data_dir, 'fashion_mnist', 'test_images.npy')
        
        if dataset_path and os.path.exists(dataset_path):
            X = np.load(dataset_path).astype(np.float32)
            # Subsample for benchmarking
            if X.shape[0] > 2000:
                np.random.seed(42)
                indices = np.random.choice(X.shape[0], 2000, replace=False)
                X = X[indices]
            return X
        else:
            pytest.skip(f"Real dataset {dataset_name} not available")
    
    @pytest.mark.gpu
    @pytest.mark.slow
    @pytest.mark.parametrize("dataset_name", ["mnist", "fashion_mnist"])
    def test_real_dataset_performance(self, dataset_name):
        """Test performance on real datasets."""
        X = self.load_real_dataset(dataset_name)
        
        print(f"\n=== {dataset_name.upper()} Dataset Benchmark ===")
        print(f"Dataset shape: {X.shape}")
        
        # Benchmark GPU implementation
        start_time = time.perf_counter()
        reducer = PaCMAPGPU(n_components=2, num_iters=100, verbose=False, random_state=42)
        X_embedded = reducer.fit_transform(X)
        gpu_time = time.perf_counter() - start_time
        
        print(f"GPU time: {gpu_time:.4f}s")
        print(f"Time per sample: {gpu_time/X.shape[0]*1000:.2f}ms")
        print(f"Embedding shape: {X_embedded.shape}")
        
        # Basic quality check
        assert X_embedded.shape == (X.shape[0], 2)
        assert np.std(X_embedded) > 0.1  # Should have reasonable spread
        
        # Compare with CPU if available
        if PACMAP_CPU_AVAILABLE:
            start_time = time.perf_counter()
            cpu_reducer = pacmap.PaCMAP(n_components=2, num_iters=100, verbose=False, random_state=42)
            X_cpu_embedded = cpu_reducer.fit_transform(X, init="pca")
            cpu_time = time.perf_counter() - start_time
            
            speedup = cpu_time / gpu_time
            print(f"CPU time: {cpu_time:.4f}s")
            print(f"Speedup: {speedup:.2f}x")
            
            # For real datasets, expect significant speedup
            if X.shape[0] > 500:
                assert speedup > 1.5, f"Expected significant GPU speedup, got {speedup:.2f}x" 