"""
Core functionality tests for PaCMAP-GPU.
"""

import pytest
import numpy as np
import torch
from unittest.mock import patch, MagicMock

from pacmap_gpu import PaCMAPGPU, check_gpu_availability


class TestPaCMAPGPUInitialization:
    """Test PaCMAPGPU class initialization."""
    
    def test_default_initialization(self):
        """Test default parameter initialization."""
        reducer = PaCMAPGPU()
        
        assert reducer.n_components == 2
        assert reducer.n_neighbors == 10
        assert reducer.MN_ratio == 0.5
        assert reducer.FP_ratio == 2.0
        assert reducer.num_iters == 450
        assert reducer.batch_size == 8192
        assert reducer.init_method == "pca"
        assert reducer.optimizer_name == "adam"
        assert reducer.lr == 0.01
        assert reducer.device == "cuda"
        assert reducer.low_dist_thres == 1.0
        assert reducer.verbose is False
        assert reducer.embedding_ is None
        assert reducer.is_fitted is False
    
    def test_custom_initialization(self):
        """Test initialization with custom parameters."""
        reducer = PaCMAPGPU(
            n_components=3,
            n_neighbors=15,
            MN_ratio=0.3,
            FP_ratio=1.5,
            num_iters=100,
            batch_size=2048,
            init_method="random",
            optimizer="adagrad",
            lr=0.005,
            device="cuda:0",
            low_dist_thres=2.0,
            random_state=123,
            verbose=True
        )
        
        assert reducer.n_components == 3
        assert reducer.n_neighbors == 15
        assert reducer.MN_ratio == 0.3
        assert reducer.FP_ratio == 1.5
        assert reducer.num_iters == 100
        assert reducer.batch_size == 2048
        assert reducer.init_method == "random"
        assert reducer.optimizer_name == "adagrad"
        assert reducer.lr == 0.005
        assert reducer.device == "cuda:0"
        assert reducer.low_dist_thres == 2.0
        assert reducer.random_state == 123
        assert reducer.verbose is True
    
    def test_random_state_setting(self):
        """Test that random state is properly set."""
        with patch('torch.manual_seed') as mock_torch, \
             patch('torch.cuda.manual_seed') as mock_cuda, \
             patch('numpy.random.seed') as mock_numpy:
            
            PaCMAPGPU(random_state=42)
            
            mock_torch.assert_called_once_with(42)
            mock_cuda.assert_called_once_with(42)
            mock_numpy.assert_called_once_with(42)


class TestPaCMAPGPUFitTransform:
    """Test the main fit_transform functionality."""
    
    @pytest.mark.gpu
    def test_basic_fit_transform(self, small_2d_data, pacmap_gpu_instance):
        """Test basic fit_transform functionality."""
        X = small_2d_data
        reducer = pacmap_gpu_instance
        
        # Should not raise any exceptions
        X_embedded = reducer.fit_transform(X)
        
        # Check output shape
        assert X_embedded.shape == (X.shape[0], reducer.n_components)
        assert isinstance(X_embedded, np.ndarray)
        
        # Check that the model is marked as fitted
        assert reducer.is_fitted is True
        assert reducer.embedding_ is not None
        assert reducer.embedding_.shape == X_embedded.shape
        np.testing.assert_array_equal(X_embedded, reducer.embedding_)
    
    @pytest.mark.gpu
    def test_different_output_dimensions(self, medium_data):
        """Test with different output dimensions."""
        X = medium_data
        
        for n_components in [2, 3, 5]:
            reducer = PaCMAPGPU(n_components=n_components, num_iters=10, verbose=False)
            X_embedded = reducer.fit_transform(X)
            
            assert X_embedded.shape == (X.shape[0], n_components)
            assert reducer.is_fitted is True
    
    def test_input_validation(self):
        """Test input validation and error handling."""
        reducer = PaCMAPGPU(num_iters=10, verbose=False)
        
        # Test invalid input shapes
        with pytest.raises((ValueError, AttributeError)):
            reducer.fit_transform(np.array([1, 2, 3]))  # 1D array
        
        # Test empty array
        with pytest.raises((ValueError, RuntimeError)):
            reducer.fit_transform(np.array([]).reshape(0, 5))
    
    @pytest.mark.gpu 
    def test_reproducibility(self, small_2d_data):
        """Test that results are reproducible with same random state."""
        X = small_2d_data
        
        reducer1 = PaCMAPGPU(random_state=42, num_iters=20, verbose=False)
        reducer2 = PaCMAPGPU(random_state=42, num_iters=20, verbose=False)
        
        X_embedded1 = reducer1.fit_transform(X)
        X_embedded2 = reducer2.fit_transform(X)
        
        # Instead of elementwise closeness, check for high correlation between embeddings
        # This is more robust to GPU non-determinism and floating-point noise
        corr = np.corrcoef(X_embedded1.ravel(), X_embedded2.ravel())[0, 1]
        assert corr > 0.95, f"Embeddings correlation too low: {corr:.4f}"
    
    @pytest.mark.gpu
    def test_different_initialization_methods(self, medium_data):
        """Test different initialization methods."""
        X = medium_data
        
        # Test PCA initialization
        reducer_pca = PaCMAPGPU(init_method="pca", num_iters=10, random_state=42, verbose=False)
        X_pca = reducer_pca.fit_transform(X)
        
        # Test random initialization
        reducer_random = PaCMAPGPU(init_method="random", num_iters=10, random_state=42, verbose=False)
        X_random = reducer_random.fit_transform(X)
        
        # Both should work and produce different results
        assert X_pca.shape == X_random.shape
        assert not np.allclose(X_pca, X_random, rtol=1e-2)
    
    @pytest.mark.gpu
    def test_different_optimizers(self, small_2d_data):
        """Test different optimizers."""
        X = small_2d_data
        
        for optimizer in ["adam", "adagrad"]:
            reducer = PaCMAPGPU(optimizer=optimizer, num_iters=10, random_state=42, verbose=False)
            X_embedded = reducer.fit_transform(X)
            
            assert X_embedded.shape == (X.shape[0], 2)
            assert reducer.is_fitted is True
    
    @pytest.mark.gpu 
    def test_high_dimensional_input(self, high_dimensional_data):
        """Test with high-dimensional input that triggers PCA preprocessing."""
        X = high_dimensional_data
        
        reducer = PaCMAPGPU(num_iters=10, verbose=False, random_state=42)
        X_embedded = reducer.fit_transform(X)
        
        assert X_embedded.shape == (X.shape[0], 2)
        assert reducer.is_fitted is True


class TestPaCMAPGPUInternalMethods:
    """Test internal methods of PaCMAPGPU."""
    
    @pytest.mark.gpu
    def test_build_neighbor_graph(self, medium_data, pacmap_gpu_instance):
        """Test neighbor graph construction."""
        X = medium_data
        reducer = pacmap_gpu_instance
        reducer.n_samples = X.shape[0]
        
        indices = reducer._build_neighbor_graph(X)
        
        # Should return tensor with correct shape
        assert isinstance(indices, torch.Tensor)
        assert indices.shape == (X.shape[0], 3 * reducer.n_neighbors)
        assert indices.dtype == torch.long
        assert indices.device.type == "cuda"
    
    @pytest.mark.gpu
    def test_prepare_pair_indices(self, pacmap_gpu_instance):
        """Test pair indices preparation."""
        reducer = pacmap_gpu_instance
        n_samples = 100
        
        # Create mock neighbor indices
        indices = torch.randint(0, n_samples, (n_samples, 3 * reducer.n_neighbors), 
                               device=reducer.device, dtype=torch.long)
        
        i_N, j_N, i_MN, j_MN, i_FP, j_FP = reducer._prepare_pair_indices(indices)
        
        # Check shapes
        expected_n_pairs = n_samples * reducer.n_neighbors
        expected_mn_pairs = n_samples * int(reducer.n_neighbors * reducer.MN_ratio)
        expected_fp_pairs = n_samples * int(reducer.n_neighbors * reducer.FP_ratio)
        
        assert len(i_N) == expected_n_pairs
        assert len(j_N) == expected_n_pairs
        assert len(i_MN) == expected_mn_pairs
        assert len(j_MN) == expected_mn_pairs
        assert len(i_FP) == expected_fp_pairs
        assert len(j_FP) == expected_fp_pairs
    
    @pytest.mark.gpu
    def test_initialize_embedding(self, medium_data, pacmap_gpu_instance):
        """Test embedding initialization."""
        X = medium_data
        reducer = pacmap_gpu_instance
        
        # Test PCA initialization
        reducer.init_method = "pca"
        Y_pca = reducer._initialize_embedding(X)
        assert Y_pca.shape == (X.shape[0], reducer.n_components)
        assert Y_pca.requires_grad is True
        assert Y_pca.device.type == "cuda"
        
        # Test random initialization
        reducer.init_method = "random"
        Y_random = reducer._initialize_embedding(X)
        assert Y_random.shape == (X.shape[0], reducer.n_components)
        assert Y_random.requires_grad is True
        assert Y_random.device.type == "cuda"
    
    @pytest.mark.gpu
    def test_create_optimizer(self, pacmap_gpu_instance):
        """Test optimizer creation."""
        reducer = pacmap_gpu_instance
        Y = torch.randn(100, 2, device=reducer.device, requires_grad=True)
        
        # Test Adam optimizer
        reducer.optimizer_name = "adam"
        optimizer_adam = reducer._create_optimizer(Y)
        assert isinstance(optimizer_adam, torch.optim.Adam)
        
        # Test Adagrad optimizer  
        reducer.optimizer_name = "adagrad"
        optimizer_adagrad = reducer._create_optimizer(Y)
        assert isinstance(optimizer_adagrad, torch.optim.Adagrad)


class TestPaCMAPGPUEdgeCases:
    """Test edge cases and error conditions."""
    
    def test_small_dataset(self):
        """Test with very small dataset."""
        X = np.random.randn(5, 3).astype(np.float32)
        reducer = PaCMAPGPU(num_iters=5, verbose=False)
        
        # Should handle small datasets gracefully
        try:
            X_embedded = reducer.fit_transform(X)
            assert X_embedded.shape == (5, 2)
        except (RuntimeError, ValueError):
            # Some errors are expected with very small datasets
            pass
    
    @pytest.mark.gpu
    def test_low_dimensional_input(self):
        """Test with input dimensionality <= output dimensionality."""
        X = np.random.randn(50, 2).astype(np.float32)
        reducer = PaCMAPGPU(n_components=3, num_iters=10, verbose=False)
        
        with pytest.warns(UserWarning):
            X_embedded = reducer.fit_transform(X)
            assert X_embedded.shape == (50, 3)
    
    def test_invalid_parameters(self):
        """Test with invalid parameters."""
        # Negative dimensions
        with pytest.raises(ValueError):
            PaCMAPGPU(n_components=-1)
        
        # Zero iterations
        reducer = PaCMAPGPU(num_iters=0)
        X = np.random.randn(10, 5).astype(np.float32)
        
        # Should handle gracefully or raise appropriate error
        try:
            reducer.fit_transform(X)
        except (ValueError, RuntimeError):
            pass  # Expected behavior
    
    @patch('torch.cuda.is_available')
    def test_no_gpu_available(self, mock_cuda_available):
        """Test behavior when GPU is not available."""
        mock_cuda_available.return_value = False
        
        reducer = PaCMAPGPU(device="cuda", verbose=False)
        X = np.random.randn(20, 5).astype(np.float32)
        
        # Should either fallback gracefully or raise informative error
        with pytest.raises((RuntimeError, ValueError)):
            reducer.fit_transform(X) 