"""
Tests for utility functions in PaCMAP-GPU.
"""

import pytest
import torch
from unittest.mock import patch, MagicMock
import sys

from pacmap_gpu.utils import (
    check_gpu_availability,
    get_gpu_info, 
    clear_gpu_memory,
    check_rapids_installation,
    verify_installation,
    get_system_info
)


class TestGPUUtilities:
    """Test GPU-related utility functions."""
    
    def test_check_gpu_availability_with_cuda(self):
        """Test GPU availability check when CUDA is available."""
        with patch('torch.cuda.is_available', return_value=True), \
             patch('torch.cuda.device_count', return_value=1):
            assert check_gpu_availability() is True
    
    def test_check_gpu_availability_no_cuda(self):
        """Test GPU availability check when CUDA is not available."""
        with patch('torch.cuda.is_available', return_value=False):
            assert check_gpu_availability() is False
    
    def test_check_gpu_availability_exception(self):
        """Test GPU availability check when exception occurs."""
        with patch('torch.cuda.is_available', side_effect=Exception("Mock error")):
            assert check_gpu_availability() is False
    
    def test_get_gpu_info_no_cuda(self):
        """Test get_gpu_info when CUDA is not available."""
        with patch('torch.cuda.is_available', return_value=False):
            info = get_gpu_info()
            assert info['available'] is False
            assert info['devices'] == []
            assert 'error' in info
    
    @pytest.mark.gpu
    def test_get_gpu_info_with_cuda(self):
        """Test get_gpu_info when CUDA is available."""
        if not torch.cuda.is_available():
            pytest.skip("CUDA not available")
            
        info = get_gpu_info()
        
        assert info['available'] is True
        assert 'devices' in info
        assert 'current_device' in info
        assert 'device_count' in info
        assert 'cuda_version' in info
        assert 'pytorch_version' in info
        
        if info['device_count'] > 0:
            device = info['devices'][0]
            required_keys = ['id', 'name', 'total_memory_gb', 'capability']
            for key in required_keys:
                assert key in device
    
    def test_get_gpu_info_exception(self):
        """Test get_gpu_info when exception occurs."""
        with patch('torch.cuda.is_available', return_value=True), \
             patch('torch.cuda.device_count', side_effect=Exception("Mock error")):
            info = get_gpu_info()
            assert info['available'] is False
            assert 'error' in info
    
    @pytest.mark.gpu
    def test_clear_gpu_memory(self):
        """Test GPU memory clearing."""
        if not torch.cuda.is_available():
            pytest.skip("CUDA not available")
            
        # Should not raise any exceptions
        clear_gpu_memory()
        clear_gpu_memory(device_id=0)
    
    def test_clear_gpu_memory_no_cuda(self):
        """Test GPU memory clearing when CUDA is not available."""
        with patch('torch.cuda.is_available', return_value=False):
            with pytest.warns(UserWarning):
                clear_gpu_memory()
    
    def test_clear_gpu_memory_exception(self):
        """Test GPU memory clearing when exception occurs."""
        with patch('torch.cuda.is_available', return_value=True), \
             patch('torch.cuda.empty_cache', side_effect=Exception("Mock error")):
            with pytest.warns(UserWarning):
                clear_gpu_memory()


class TestRAPIDSInstallation:
    """Test RAPIDS installation checking."""
    
    def test_check_rapids_installation_all_available(self):
        """Test when all RAPIDS components are available."""
        mock_modules = {
            'cupy': MagicMock(__version__='12.0.0'),
            'cuml': MagicMock(__version__='24.0.0'),
            'cudf': MagicMock(__version__='24.0.0'),
            'faiss': MagicMock(__version__='1.7.0')
        }
        
        def mock_import(name):
            if name in mock_modules:
                return mock_modules[name]
            raise ImportError(f"No module named '{name}'")
        
        with patch('importlib.import_module', side_effect=mock_import):
            results = check_rapids_installation()
            
            for component in ['cupy', 'cuml', 'cudf', 'faiss-gpu']:
                assert results[component]['installed'] is True
                assert 'version' in results[component]
    
    def test_check_rapids_installation_none_available(self):
        """Test when no RAPIDS components are available."""
        def mock_import(name):
            raise ImportError(f"No module named '{name}'")
        
        with patch('importlib.import_module', side_effect=mock_import):
            results = check_rapids_installation()
            
            for component in ['cupy', 'cuml', 'cudf', 'faiss-gpu']:
                assert results[component]['installed'] is False
                assert 'error' in results[component]
    
    def test_check_rapids_installation_partial(self):
        """Test when some RAPIDS components are available."""
        def mock_import(name):
            if name == 'cupy':
                return MagicMock(__version__='12.0.0')
            raise ImportError(f"No module named '{name}'")
        
        with patch('importlib.import_module', side_effect=mock_import):
            results = check_rapids_installation()
            
            assert results['cupy']['installed'] is True
            assert results['cuml']['installed'] is False
            assert results['cudf']['installed'] is False
            assert results['faiss-gpu']['installed'] is False


class TestInstallationVerification:
    """Test comprehensive installation verification."""
    
    def test_verify_installation_basic_structure(self):
        """Test that verify_installation returns proper structure."""
        results = verify_installation()
        
        required_keys = [
            'python_version', 'python_compatible', 'gpu_info', 
            'rapids_info', 'core_dependencies', 'overall_status'
        ]
        
        for key in required_keys:
            assert key in results
        
        assert isinstance(results['python_compatible'], bool)
        assert isinstance(results['gpu_info'], dict)
        assert isinstance(results['rapids_info'], dict)
        assert isinstance(results['core_dependencies'], dict)
        assert results['overall_status'] in ['excellent', 'good', 'basic', 'poor', 'unknown']
    
    def test_verify_installation_python_compatibility(self):
        """Test Python version compatibility checking."""
        results = verify_installation()
        
        # Should correctly identify Python 3.10+ compatibility
        expected_compatible = sys.version_info >= (3, 10)
        assert results['python_compatible'] == expected_compatible
    
    def test_verify_installation_core_dependencies(self):
        """Test core dependencies checking."""
        results = verify_installation()
        
        expected_deps = ['torch', 'numpy', 'scipy', 'sklearn', 'pandas', 'matplotlib']
        for dep in expected_deps:
            assert dep in results['core_dependencies']
            dep_info = results['core_dependencies'][dep]
            assert 'installed' in dep_info
            if dep_info['installed']:
                assert 'version' in dep_info
            else:
                assert 'error' in dep_info
    
    @patch('sys.version_info', (3, 9))  # Python 3.9
    def test_verify_installation_old_python(self):
        """Test with old Python version."""
        results = verify_installation()
        assert results['python_compatible'] is False
        assert results['overall_status'] in ['basic', 'poor']
    
    def test_overall_status_logic(self):
        """Test overall status determination logic."""
        # Mock a perfect installation
        with patch('pacmap_gpu.utils.get_gpu_info', return_value={'available': True}), \
             patch('pacmap_gpu.utils.check_rapids_installation', return_value={
                 'cupy': {'installed': True}, 'cuml': {'installed': True},
                 'cudf': {'installed': True}, 'faiss-gpu': {'installed': True}
             }), \
             patch('sys.version_info', (3, 10)):
            
            # Mock all core dependencies as available
            def mock_import(name):
                return MagicMock(__version__='1.0.0')
            
            with patch('importlib.import_module', side_effect=mock_import):
                results = verify_installation()
                assert results['overall_status'] == 'excellent'


class TestSystemInfo:
    """Test system information gathering."""
    
    def test_get_system_info_structure(self):
        """Test that get_system_info returns proper structure."""
        info = get_system_info()
        
        required_keys = ['platform', 'python', 'environment']
        for key in required_keys:
            assert key in info
        
        # Check platform info
        platform_keys = ['system', 'release', 'version', 'machine', 'processor']
        for key in platform_keys:
            assert key in info['platform']
        
        # Check python info
        python_keys = ['version', 'executable', 'platform']
        for key in python_keys:
            assert key in info['python']
        
        # Check environment info
        env_keys = ['cuda_home', 'cuda_path', 'ld_library_path', 'path']
        for key in env_keys:
            assert key in info['environment']
    
    def test_get_system_info_values(self):
        """Test that get_system_info returns reasonable values."""
        info = get_system_info()
        
        # Python version should be current version
        assert info['python']['version'] == sys.version
        assert info['python']['executable'] == sys.executable
        
        # Platform system should be valid
        assert info['platform']['system'] in ['Windows', 'Linux', 'Darwin']


class TestReportFunctions:
    """Test report generation functions."""
    
    def test_print_installation_report(self, capsys):
        """Test that print_installation_report produces output."""
        from pacmap_gpu.utils import print_installation_report
        
        print_installation_report()
        
        captured = capsys.readouterr()
        assert "PaCMAP-GPU Installation Verification Report" in captured.out
        assert "Python Version:" in captured.out
        assert "GPU" in captured.out or "CUDA" in captured.out
        assert "Overall Status:" in captured.out
    
    def test_main_check_installation(self, capsys):
        """Test main installation checking function."""
        from pacmap_gpu.utils import main_check_installation
        
        main_check_installation()
        
        captured = capsys.readouterr()
        assert len(captured.out) > 0  # Should produce some output
        assert "PaCMAP-GPU Installation Verification Report" in captured.out 